@@ -88,6 +88,44 @@ def set_seed(seed, n_gpu):
         torch.cuda.manual_seed_all(seed)
 
 
+class Prefetcher:
+    def __init__(self, dataloader, stream):
+        self.dataloader = dataloader
+        self.stream = torch.cuda.Stream()
+
+    def __iter__(self):
+        self.iter = iter(self.dataloader)
+        self.preload()
+        return self
+
+    def preload(self):
+        try:
+            self.next = next(self.iter)
+        except StopIteration:
+            self.next = None
+            return
+        with torch.cuda.stream(self.stream):
+            next_list = list()
+            for v in self.next:
+                if type(v) == torch.Tensor:
+                    next_list.append(v.cuda(non_blocking=True))
+                else:
+                    next_list.append(v)
+            self.next = tuple(next_list)
+            
+    def __next__(self):
+        torch.cuda.current_stream().wait_stream(self.stream)
+        if self.next is not None:
+            result = self.next
+            self.preload()
+            return result
+        else:
+            raise StopIteration
+
+    def __len__(self):
+        return len(self.dataloader)
+
+
 class TrainerCallback:
     def __init__(self):
         pass
@@ -250,6 +288,7 @@ class Trainer:
         self.eval_batch_size = self.per_gpu_eval_batch_size * max(1, self.n_gpu)
         if self.fp16:
             apex.amp.register_half_function(torch, "einsum")
+        self.stream = torch.cuda.Stream()
 
     def set_model(self):
         self.model = self.callback.load_model()
@@ -287,7 +326,7 @@ class Trainer:
                 train_dataset = torch.utils.data.Subset(train_dataset, list(range(int(len(train_dataset) * self.dataset_ratio))))
             self.train_dataset = train_dataset
             train_sampler = RandomSampler(self.train_dataset) if self.local_rank == -1 else DistributedSampler(self.train_dataset)
-            self.train_dataloader = DataLoader(self.train_dataset, sampler=train_sampler, batch_size=self.train_batch_size, collate_fn=train_fn, num_workers=self.num_workers)
+            self.train_dataloader = Prefetcher(DataLoader(self.train_dataset, sampler=train_sampler, batch_size=self.train_batch_size, collate_fn=train_fn, num_workers=self.num_workers), self.stream)
             self.t_total = len(self.train_dataloader) // self.gradient_accumulation_steps * self.epochs
             self.scheduler = get_linear_schedule_with_warmup(self.optimizer, num_warmup_steps=int(self.t_total * self.warmup_ratio), num_training_steps=self.t_total)
         if dev_dataset:
@@ -295,13 +334,13 @@ class Trainer:
                 dev_dataset = torch.utils.data.Subset(dev_dataset, list(range(int(len(dev_dataset) * self.dataset_ratio))))
             self.dev_dataset = dev_dataset
             dev_sampler = SequentialSampler(self.dev_dataset) if self.local_rank == -1 else DistributedSampler(self.dev_dataset)
-            self.dev_dataloader = DataLoader(self.dev_dataset, sampler=dev_sampler, batch_size=self.eval_batch_size, collate_fn=dev_fn, num_workers=self.num_workers)
+            self.dev_dataloader = Prefetcher(DataLoader(self.dev_dataset, sampler=dev_sampler, batch_size=self.eval_batch_size, collate_fn=dev_fn, num_workers=self.num_workers), self.stream)
         if test_dataset:
             if self.dataset_ratio < 1:
                 test_dataset = torch.utils.data.Subset(test_dataset, list(range(int(len(test_dataset) * self.dataset_ratio))))
             self.test_dataset = test_dataset
             test_sampler = SequentialSampler(self.test_dataset) if self.local_rank == -1 else DistributedSampler(self.test_dataset)
-            self.test_dataloader = DataLoader(self.test_dataset, sampler=test_sampler, batch_size=self.eval_batch_size, collate_fn=test_fn, num_workers=self.num_workers)
+            self.test_dataloader = Prefetcher(DataLoader(self.test_dataset, sampler=test_sampler, batch_size=self.eval_batch_size, collate_fn=test_fn, num_workers=self.num_workers), self.stream)
 
     def restore_checkpoint(self, path, ignore_progress=False):
         if self.no_save:
